import mmh3
import numpy as np


def n_feature_hash(feature, dims, seeds):
    """N-hot-encoded feature hashing.

    Args:
        feature (str): Target feature represented as string.
        dims (list of int): Number of dimensions for each hash value.
        seeds (list of float): Seed of each hash function (mmh3).

    Returns:
        numpy 1d array: n-hot-encoded feature vector for `s`.

    """
    vec = np.zeros(sum(dims))
    offset = 0

    for seed, dim in zip(seeds, dims):
        vec[offset:(offset + dim)] = feature_hash(feature, dim, seed)
        offset += dim

    return vec


def feature_hash(feature, dim, seed=123):
    """Feature hashing.

    Args:
        feature (str): Target feature represented as string.
        dim (int): Number of dimensions for a hash value.
        seed (float): Seed of a MurmurHash3 hash function.

    Returns:
        numpy 1d array: one-hot-encoded feature vector for `s`.

    """
    vec = np.zeros(dim)
    i = mmh3.hash(feature, seed) % dim
    vec[i] = 1
    return vec


def multiple_feature_hash(feature, dim, seed=123):
    """Feature hashing using multiple hash function.
    This technique is effective to prevent collisions.

    Args:
        feature (str): Target feature represented as string.
        dim (int): Number of dimensions for a hash value.
        seed (float): Seed of a MurmurHash3 hash function.

    Returns:
        numpy 1d array: one-hot-encoded feature vector for `s`.

    """
    vec = np.zeros(dim)
    i = mmh3.hash(feature, seed) % dim
    vec[i] = 1 if mmh3.hash(feature) % 2 else -1
    return vec



from flurs.base import BaseModel

import numpy as np


class MatrixFactorization(BaseModel):

    """Incremental Matrix Factorization

    J. Vinagre et al.
    "Fast Incremental Matrix Factorization for Recommendation with Positive-Only Feedback"
    In Proceedings of UMAP 2014, pages 459-470, July 2014.

    """

    def __init__(self, k=40, l2_reg=.01, learn_rate=.003):
        self.k = k
        self.l2_reg_u = l2_reg
        self.l2_reg_i = l2_reg
        self.learn_rate = learn_rate

        self.init_params()

    def init_params(self):
        self.Q = np.array([])

    def update_params(self, ua, ia, value):
        u_vec = self.users[ua]['vec']
        i_vec = self.Q[ia]

        err = value - np.inner(u_vec, i_vec)

        grad = -2. * (err * i_vec - self.l2_reg_u * u_vec)
        next_u_vec = u_vec - self.learn_rate * grad

        grad = -2. * (err * u_vec - self.l2_reg_i * i_vec)
        next_i_vec = i_vec - self.learn_rate * grad

        self.users[ua]['vec'] = next_u_vec
        self.Q[ia] = next_i_vec



from flurs.base import FeatureRecommenderMixin
from flurs.model.fm import FactorizationMachine

import numpy as np
import scipy.sparse as sp
from sklearn.utils.extmath import safe_sparse_dot


class FMRecommender(FactorizationMachine, FeatureRecommenderMixin):

    def init_recommender(self, is_static=False, use_index=False):
        super().init_recommender()
        self.is_static = is_static
        self.use_index = use_index

    def add_user(self, user):
        super().add_user(user)

        if self.use_index:
            n_user = self.n_user - 1

            # insert new user's row for the parameters
            self.w = np.concatenate((self.w[:n_user],
                                     np.array([0.]),
                                     self.w[n_user:]))

            self.prev_w = np.concatenate((self.prev_w[:n_user],
                                          np.array([0.]),
                                          self.prev_w[n_user:]))

            rand_row = np.random.normal(0., 0.1, (1, self.k))

            self.V = np.concatenate((self.V[:n_user],
                                     rand_row,
                                     self.V[n_user:]))

            self.prev_V = np.concatenate((self.prev_V[:n_user],
                                          rand_row,
                                          self.prev_V[n_user:]))

            self.p += 1

    def add_item(self, item):
        super().add_item(item)

        n_item = self.n_item - 1

        # update the item matrix for all items
        i_vec = item.encode(dim=(n_item + 1),
                            index=self.use_index,
                            feature=True,
                            vertical=True)
        sp_i_vec = sp.csr_matrix(i_vec)

        if self.i_mat.size == 0:
            self.i_mat = sp_i_vec
        elif self.use_index:
            self.i_mat = sp.vstack((self.i_mat[:n_item],
                                    np.zeros((1, n_item)),
                                    self.i_mat[n_item:]))
            self.i_mat = sp.csr_matrix(sp.hstack((self.i_mat, sp_i_vec)))
        else:
            self.i_mat = sp.csr_matrix(sp.hstack((self.i_mat, sp_i_vec)))

        if self.use_index:

            # insert new item's row for the parameters
            h = self.n_user + n_item

            self.w = np.concatenate((self.w[:h],
                                     np.array([0.]),
                                     self.w[h:]))

            self.prev_w = np.concatenate((self.prev_w[:h],
                                          np.array([0.]),
                                          self.prev_w[h:]))

            rand_row = np.random.normal(0., 0.1, (1, self.k))

            self.V = np.concatenate((self.V[:h],
                                     rand_row,
                                     self.V[h:]))

            self.prev_V = np.concatenate((self.prev_V[:h],
                                          rand_row,
                                          self.prev_V[h:]))
            self.p += 1

    def update(self, e, is_batch_train=False):
        # static baseline; w/o updating the model
        if not is_batch_train and self.is_static:
            return

        x = e.encode(index=self.use_index,
                     n_user=self.n_user, n_item=self.n_item)

        self.update_params(x, e.value)

    def score(self, user, candidates, context):
        # i_mat is (n_item_context, n_item) for all possible items
        # extract only target items
        i_mat = self.i_mat[:, candidates]

        n_target = len(candidates)

        u_vec = user.encode(dim=self.n_user,
                            index=self.use_index,
                            feature=True,
                            vertical=True)
        u_vec = np.concatenate((u_vec, np.array([context]).T))
        u_mat = sp.csr_matrix(np.repeat(u_vec, n_target, axis=1))

        mat = sp.vstack((u_mat, i_mat))

        # Matrix A and B should be dense (numpy array; rather than scipy CSR matrix) because V is dense.
        V = sp.csr_matrix(self.V)
        A = safe_sparse_dot(V.T, mat)
        A.data[:] = A.data ** 2

        sq_mat = mat.copy()
        sq_mat.data[:] = sq_mat.data ** 2
        sq_V = V.copy()
        sq_V.data[:] = sq_V.data ** 2
        B = safe_sparse_dot(sq_V.T, sq_mat)

        interaction = (A - B).sum(axis=0)
        interaction /= 2.  # (1, n_item); numpy matrix form

        pred = self.w0 + safe_sparse_dot(self.w, mat, dense_output=True) + interaction

        return np.abs(1. - np.ravel(pred))

    def recommend(self, user, candidates, context):
        scores = self.score(user, candidates, context)
        return self.scores2recos(scores, candidates)



import numpy as np


def count_true_positive(truth, recommend):
    """Count number of true positives from given sets of samples.

    Args:
        truth (numpy 1d array): Set of truth samples.
        recommend (numpy 1d array): Ordered set of recommended samples.

    Returns:
        int: Number of true positives.

    """
    tp = 0
    for r in recommend:
        if r in truth:
            tp += 1
    return tp


def recall(truth, recommend, k):
    """Recall@k.

    Args:
        truth (numpy 1d array): Set of truth samples.
        recommend (numpy 1d array): Ordered set of recommended samples.
        k (int): Top-k items in `recommend` will be recommended.

    Returns:
        float: Recall@k.

    """
    return count_true_positive(truth, recommend[:k]) / truth.size


def precision(truth, recommend, k):
    """Precision@k.

    Args:
        truth (numpy 1d array): Set of truth samples.
        recommend (numpy 1d array): Ordered set of recommended samples.
        k (int): Top-k items in `recommend` will be recommended.

    Returns:
        float: Precision@k.

    """
    return count_true_positive(truth, recommend[:k]) / k


def mean_average_precision(truth, recommend):
    """Mean Average Precision (MAP).

    Args:
        truth (numpy 1d array): Set of truth samples.
        recommend (numpy 1d array): Ordered set of recommended samples.

    Returns:
        float: MAP.

    """
    tp = accum = 0
    for n in range(recommend.size):
        if recommend[n] in truth:
            tp += 1
            accum += (tp / (n + 1))
    return accum / truth.size


def auc(truth, recommend):
    """Area under the ROC curve (AUC).

    Args:
        truth (numpy 1d array): Set of truth samples.
        recommend (numpy 1d array): Ordered set of recommended samples.

    Returns:
        float: AUC.

    """
    tp = correct = 0
    for r in recommend:
        if r in truth:
            # keep track number of true positives placed before
            tp += 1
        else:
            correct += tp
    # number of all possible tp-fp pairs
    pairs = tp * (recommend.size - tp)
    return correct / pairs


def mrr(truth, recommend):
    """Mean Reciprocal Rank (MRR).

    Args:
        truth (numpy 1d array): Set of truth samples.
        recommend (numpy 1d array): Ordered set of recommended samples.

    Returns:
        float: MRR.

    """
    for n in range(recommend.size):
        if recommend[n] in truth:
            return 1 / (n + 1)
    return 0


def mpr(truth, recommend):
    """Mean Percentile Rank (MPR).

    Args:
        truth (numpy 1d array): Set of truth samples.
        recommend (numpy 1d array): Ordered set of recommended samples.

    Returns:
        float: MPR.

    """
    accum = 0
    n_recommend = recommend.size
    for t in truth:
        r = np.where(recommend == t)[0][0] / n_recommend
        accum += r
    return accum * 100 / truth.size


def ndcg(truth, recommend, k):
    """Normalized Discounted Cumulative Grain (NDCG).

    Args:
        truth (numpy 1d array): Set of truth samples.
        recommend (numpy 1d array): Ordered set of recommended samples.
        k (int): Top-k items in `recommend` will be recommended.

    Returns:
        float: NDCG.

    """
    dcg = idcg = 0
    for n in range(k):
        d = 1 / np.log2(n + 2)
        if recommend[n] in truth:
            dcg += d
        idcg += d
    return dcg / idcg



from flurs.base import RecommenderMixin
from flurs.model.user_knn import UserKNN

import numpy as np


class UserKNNRecommender(UserKNN, RecommenderMixin):

    def init_recommender(self):
        super().init_recommender()

    def insert_row(self, mat, n_col):
        row = np.zeros((1, n_col))
        if mat.size == 0:
            mat = row
        else:
            mat = np.concatenate((mat, row))
        return mat

    def insert_col(self, mat, n_row):
        col = np.zeros((n_row, 1))
        if mat.size == 0:
            mat = col
        else:
            mat = np.concatenate((mat, col), axis=1)
        return mat

    def add_user(self, user):
        super().add_user(user)

        self.R = self.insert_row(self.R, self.n_item)

        self.S = self.insert_row(self.S, self.n_user - 1)
        self.S = self.insert_col(self.S, self.n_user)

        self.B = self.insert_row(self.B, self.n_user - 1)
        self.B = self.insert_col(self.B, self.n_user)

        self.C = self.insert_row(self.C, self.n_user - 1)
        self.C = self.insert_col(self.C, self.n_user)

        self.D = self.insert_row(self.D, self.n_user - 1)
        self.D = self.insert_col(self.D, self.n_user)

        # keep track how many times each user interacted with items
        # to compute user's mean
        self.users[user.index]['count'] = 0

        # current average rating of the user
        self.users[user.index]['mean'] = 0.

    def add_item(self, item):
        super().add_item(item)
        self.R = self.insert_col(self.R, self.n_user)

    def update(self, e, is_batch_train=False):
        self.update_params(e.user.index, e.item.index, e.value)

    def score(self, user, candidates):
        ua = user.index

        # find k nearest neightbors
        top_uys = np.argsort(self.S[ua, :])[::-1][:self.k]

        pred = np.ones(len(candidates)) * self.users[ua]['mean']
        for pi, ii in enumerate(candidates):
            denom = numer = 0.
            for uy in top_uys:
                numer += ((self.R[uy, ii] - self.users[uy]['mean']) * self.S[ua, uy])
                denom += self.S[ua, uy]
            if denom != 0.:
                pred[pi] += (numer / denom)

        return np.abs(pred)

    def recommend(self, user, candidates):
        scores = self.score(user, candidates)
        return self.scores2recos(scores, candidates, rev=True)



from flurs.base import BaseModel

import numpy as np


class BPRMF(BaseModel):

    """Incremental Matrix Factorization with BPR optimization

    S. Rendle et al.
    "BPR: Bayesian Personalized Ranking from Implicit Feedback"
    In Proceedings of UAI 2009, pages 452-461, June 2009.

    """

    def __init__(self, k=40, l2_reg=.01, learn_rate=.003):
        self.k = k
        self.l2_reg_u = l2_reg
        self.l2_reg_i = l2_reg  # positive items: i
        self.l2_reg_j = l2_reg  # negative items: j
        self.learn_rate = learn_rate

        self.init_params()

    def init_params(self):
        self.Q = np.array([])

    def update_params(self, ua, ia):

        u_vec = self.users[ua]['vec']
        i_vec = self.Q[ia]
        x_ui = np.inner(u_vec, i_vec)

        unobserved = list(set(range(self.n_item)) - self.users[ua]['known_items'])

        # choose one negative (i.e., unobserved) sample
        j = np.random.choice(unobserved)

        j_vec = self.Q[j]

        x_uj = np.inner(u_vec, j_vec)
        x_uij = x_ui - x_uj
        sigmoid = np.e ** (-x_uij) / (1 + np.e ** (-x_uij))

        grad = i_vec - j_vec
        next_u_vec = u_vec + self.learn_rate * (sigmoid * grad + self.l2_reg_u * u_vec)

        grad = u_vec
        next_i_vec = i_vec + self.learn_rate * (sigmoid * grad + self.l2_reg_i * i_vec)

        grad = -u_vec
        next_j_vec = j_vec + self.learn_rate * (sigmoid * grad + self.l2_reg_j * j_vec)

        self.users[ua]['vec'] = next_u_vec
        self.Q[ia] = next_i_vec
        self.Q[j] = next_j_vec



from abc import ABCMeta, abstractmethod

import numpy as np
import scipy.sparse as sp
from sklearn.utils.extmath import safe_sparse_dot


class BaseProjection:

    """Base class for projection of feature vectors
    """
    __metaclass__ = ABCMeta

    @abstractmethod
    def __init__(self, k, p):
        """Initialize projection matrices.

        Args:
            k (int): Number of reduced dimensions (i.e. rows of projection mat.).
            p (int): Number of input dimensions (i.e. cols of projection mat.).
                p will be increased in the future due to new user/item/context insersion.

        """
        pass

    @abstractmethod
    def insert_proj_col(self, offset):
        """Insert a new column for a projection matrix.

        Args:
            offset (int): Index of the inserted column.

        """
        pass

    @abstractmethod
    def reduce(self, Y):
        """Make projection for an input matrix.

        Args:
            Y (numpy array; (p, n)): Input p-by-n matrix projected to a k-by-n matrix.

        Returns:
            numpy array; (k, n): Projected matrix.

        """
        return


class Raw(BaseProjection):

    def __init__(self, k, p):
        # k == p
        self.E = np.identity(p)

    def insert_proj_col(self, offset):
        pass

    def reduce(self, Y):
        return safe_sparse_dot(self.E, Y)


class RandomProjection(BaseProjection):

    def __init__(self, k, p):
        self.k = k
        self.R = sp.csr_matrix(self.__create_proj_mat((k, p)))

    def insert_proj_col(self, offset):
        col = self.__create_proj_mat((self.k, 1))
        R = self.R.toarray()
        self.R = sp.csr_matrix(np.concatenate((R[:, :offset], col, R[:, offset:]), axis=1))

    def reduce(self, Y):
        return safe_sparse_dot(self.R, Y)

    def __create_proj_mat(self, size):
        """Create a random projection matrix

        [1] D. Achlioptas. Database-friendly random projections: Johnson-Lindenstrauss with binary coins.
        [2] P. Li, et al. Very sparse random projections.

        http://scikit-learn.org/stable/modules/random_projection.html#sparse-random-projection
        """

        # [1]
        # return np.random.choice([-np.sqrt(3), 0, np.sqrt(3)], size=size, p=[1 / 6, 2 / 3, 1 / 6])

        # [2]
        s = 1 / 0.2
        return np.random.choice([-np.sqrt(s / self.k), 0, np.sqrt(s / self.k)],
                                size=size,
                                p=[1 / (2 * s), 1 - 1 / s, 1 / (2 * s)])


class RandomMaclaurinProjection(BaseProjection):

    def __init__(self, k, p):
        self.k = k

        self.W1 = np.random.choice([1, -1], size=(k, p))
        self.W2 = np.random.choice([1, -1], size=(k, p))

    def insert_proj_col(self, offset):
        col = np.random.choice([1, -1], size=(self.k, 1))
        self.W1 = np.concatenate((self.W1[:, :offset], col, self.W1[:, offset:]), axis=1)

        col = np.random.choice([1, -1], size=(self.k, 1))
        self.W2 = np.concatenate((self.W2[:, :offset], col, self.W2[:, offset:]), axis=1)

    def reduce(self, Y):
        return safe_sparse_dot(self.W1, Y) * safe_sparse_dot(self.W2, Y) / np.sqrt(self.k)


class TensorSketchProjection(BaseProjection):

    def __init__(self, k, p):
        self.k = k

        self.h1 = np.random.choice(range(k), size=p)
        self.h2 = np.random.choice(range(k), size=p)
        self.h1_indices = [np.where(self.h1 == j)[0] for j in range(k)]
        self.h2_indices = [np.where(self.h2 == j)[0] for j in range(k)]
        self.s1 = np.random.choice([1, -1], size=p)
        self.s2 = np.random.choice([1, -1], size=p)

    def insert_proj_col(self, offset):
        self.h1 = np.concatenate((self.h1[:offset],
                                  np.random.choice(range(self.k), (1, )),
                                  self.h1[offset:]))
        self.h2 = np.concatenate((self.h2[:offset],
                                  np.random.choice(range(self.k), (1, )),
                                  self.h2[offset:]))
        self.h1_indices = [np.where(self.h1 == j)[0] for j in range(self.k)]
        self.h2_indices = [np.where(self.h2 == j)[0] for j in range(self.k)]
        self.s1 = np.concatenate((self.s1[:offset], np.random.choice([1, -1], (1, )), self.s1[offset:]))
        self.s2 = np.concatenate((self.s2[:offset], np.random.choice([1, -1], (1, )), self.s2[offset:]))

    def reduce(self, Y):
        if sp.isspmatrix(Y):
            Y = Y.toarray()

        sketch1, sketch2 = self.__sketch(Y)

        return np.real(np.fft.ifft(np.fft.fft(sketch1, axis=0) * np.fft.fft(sketch2, axis=0), axis=0))

    def __sketch(self, X):
        sketch1 = np.array([np.sum(np.array([self.s1[idx]]).T * X[idx], axis=0) for idx in self.h1_indices])
        sketch2 = np.array([np.sum(np.array([self.s2[idx]]).T * X[idx], axis=0) for idx in self.h2_indices])

        return sketch1, sketch2



from flurs.base import FeatureRecommenderMixin

import time
import numpy as np

from collections import deque
from logging import getLogger, StreamHandler, Formatter, DEBUG
logger = getLogger(__name__)
handler = StreamHandler()
handler.setFormatter(Formatter('[%(process)d] %(message)s'))
handler.setLevel(DEBUG)
logger.setLevel(DEBUG)
logger.addHandler(handler)


class Evaluator:

    """Base class for experimentation of the incremental models with positive-only feedback.
    """

    def __init__(self, recommender, can_repeat=True, maxlen=None):
        """Set/initialize parameters.

        Args:
            recommender (Recommender): Instance of a recommender which has been initialized.
            can_repeat (boolean): Choose whether the same item can be repeatedly interacted by the same user.
            maxlen (int): Size of an item buffer which stores most recently observed items.

        """
        self.rec = recommender
        self.is_feature_rec = issubclass(recommender.__class__, FeatureRecommenderMixin)

        self.can_repeat = can_repeat

        # create a ring buffer
        # save items which are observed in most recent `maxlen` events
        self.item_buffer = deque(maxlen=maxlen)

        # initialize models and user/item information
        self.rec.init_params()

    def fit(self, train_events, test_events, n_epoch=1):
        """Train a model using the first 30% positive events to avoid cold-start.

        Evaluation of this batch training is done by using the next 20% positive events.
        After the batch SGD training, the models are incrementally updated by using the 20% test events.

        Args:
            train_events (list of Event): Positive training events (0-30%).
            test_events (list of Event): Test events (30-50%).
            n_epoch (int): Number of epochs for the batch training.

        """
        self.rec.init_params()

        # make initial status for batch training
        for e in train_events:
            self.__validate(e)
            self.rec.users[e.user.index]['known_items'].add(e.item.index)
            self.item_buffer.append(e.item.index)

        # for batch evaluation, temporarily save new users info
        for e in test_events:
            self.__validate(e)
            self.item_buffer.append(e.item.index)

        self.__batch_update(train_events, test_events, n_epoch)

        # batch test events are considered as a new observations;
        # the model is incrementally updated based on them before the incremental evaluation step
        for e in test_events:
            self.rec.users[e.user.index]['known_items'].add(e.item.index)
            self.rec.update(e)

    def evaluate(self, test_events):
        """Iterate recommend/update procedure and compute incremental recall.

        Args:
            test_events (list of Event): Positive test events.

        Returns:
            list of tuples: (rank, recommend time, update time)

        """
        for i, e in enumerate(test_events):
            self.__validate(e)

            # target items (all or unobserved depending on a detaset)
            unobserved = set(self.item_buffer)
            if not self.can_repeat:
                unobserved -= self.rec.users[e.user.index]['known_items']
                # * item i interacted by user u must be in the recommendation candidate
                unobserved.add(e.item.index)
            candidates = np.asarray(list(unobserved))

            # make top-{at} recommendation for the 1001 items
            start = time.clock()
            recos, scores = self.__recommend(e, candidates)
            recommend_time = (time.clock() - start)

            rank = np.where(recos == e.item.index)[0][0]

            # Step 2: update the model with the observed event
            self.rec.users[e.user.index]['known_items'].add(e.item.index)
            start = time.clock()
            self.rec.update(e)
            update_time = (time.clock() - start)

            self.item_buffer.append(e.item.index)

            # (top-1 score, where the correct item is ranked, rec time, update time)
            yield scores[0], rank, recommend_time, update_time

    def __recommend(self, e, candidates):
        if self.is_feature_rec:
            return self.rec.recommend(e.user, candidates, e.context)
        else:
            return self.rec.recommend(e.user, candidates)

    def __validate(self, e):
        self.__validate_user(e)
        self.__validate_item(e)

    def __validate_user(self, e):
        if self.rec.is_new_user(e.user.index):
            self.rec.add_user(e.user)

    def __validate_item(self, e):
        if self.rec.is_new_item(e.item.index):
            self.rec.add_item(e.item)

    def __batch_update(self, train_events, test_events, n_epoch):
        """Batch update called by the fitting method.

        Args:
            train_events (list of Event): Positive training events (0-20%).
            test_events (list of Event): Test events (20-30%).
            n_epoch (int): Number of epochs for the batch training.

        """
        for epoch in range(n_epoch):
            # SGD requires us to shuffle events in each iteration
            # * if n_epoch == 1
            #   => shuffle is not required because it is a deterministic training (i.e. matrix sketching)
            if n_epoch != 1:
                np.random.shuffle(train_events)

            # 20%: update models
            for e in train_events:
                self.rec.update(e, is_batch_train=True)

            # 10%: evaluate the current model
            MPR = self.__batch_evaluate(test_events)
            logger.debug('epoch %2d: MPR = %f' % (epoch + 1, MPR))

    def __batch_evaluate(self, test_events):
        """Evaluate the current model by using the given test events.

        Args:
            test_events (list of Event): Current model is evaluated by these events.

        Returns:
            float: Mean Percentile Rank for the test set.

        """
        percentiles = np.zeros(len(test_events))

        all_items = set(self.item_buffer)
        for i, e in enumerate(test_events):

            # check if the data allows users to interact the same items repeatedly
            unobserved = all_items
            if not self.can_repeat:
                # make recommendation for all unobserved items
                unobserved -= self.rec.users[e.user.index]['known_items']
                # true item itself must be in the recommendation candidates
                unobserved.add(e.item.index)

            candidates = np.asarray(list(unobserved))
            recos, scores = self.__recommend(e, candidates)

            pos = np.where(recos == e.item.index)[0][0]
            percentiles[i] = pos / (len(recos) - 1) * 100

        return np.mean(percentiles)



from flurs.base import RecommenderMixin
from flurs.model.bprmf import BPRMF

import numpy as np


class BPRMFRecommender(BPRMF, RecommenderMixin):

    def init_recommender(self):
        super().init_recommender()

    def add_user(self, user):
        super().add_user(user)
        self.users[user.index]['vec'] = np.random.normal(0., 0.1, self.k)

    def add_item(self, item):
        super().add_item(item)
        i_vec = np.random.normal(0., 0.1, (1, self.k))
        if self.Q.size == 0:
            self.Q = i_vec
        else:
            self.Q = np.concatenate((self.Q, i_vec))

    def update(self, e, is_batch_train=False):
        self.update_params(e.user.index, e.item.index)

    def score(self, user, candidates):
        pred = np.dot(self.users[user.index]['vec'],
                      self.Q[candidates, :].T)
        return pred.flatten()

    def recommend(self, user, candidates):
        scores = self.score(user, candidates)
        return self.scores2recos(scores, candidates, rev=True)



from flurs.base import BaseModel

import numpy as np
import scipy.sparse as sp

from logging import getLogger, StreamHandler, Formatter, DEBUG
logger = getLogger(__name__)
handler = StreamHandler()
handler.setFormatter(Formatter('[%(process)d] %(message)s'))
handler.setLevel(DEBUG)
logger.setLevel(DEBUG)
logger.addHandler(handler)


class FactorizationMachine(BaseModel):

    """Incremental Factorization Machines

    T. Kitazawa.
    "Incremental Factorization Machines for Persistently Cold-Starting Online Item Recommendation"
    arXiv:1607.02858 [cs.LG], July 2016.

    """

    def __init__(self,
                 p,
                 k=40,
                 l2_reg_w0=2.,
                 l2_reg_w=8.,
                 l2_reg_V=16.,
                 learn_rate=.004):

        # number of dimensions of input vectors
        self.p = p

        self.k = k
        self.l2_reg_w0 = l2_reg_w0
        self.l2_reg_w = l2_reg_w
        self.l2_reg_V = np.ones(k) * l2_reg_V
        self.learn_rate = learn_rate

        self.init_params()

    def init_params(self):
        self.i_mat = sp.csr_matrix([])

        # initial parameters
        self.w0 = 0.
        self.w = np.zeros(self.p)
        self.V = np.random.normal(0., 0.1, (self.p, self.k))

        # to keep the last parameters for adaptive regularization
        self.prev_w0 = self.w0
        self.prev_w = self.w.copy()
        self.prev_V = self.V.copy()

    def update_reg(self, x, err):
        x_vec = np.array([x]).T  # p x 1

        # update regularization parameters
        coeff = 4. * self.learn_rate * err * self.learn_rate

        self.l2_reg_w0 = max(0., self.l2_reg_w0 + coeff * self.prev_w0)
        self.l2_reg_w = max(0., self.l2_reg_w + coeff * np.inner(x, self.prev_w))

        if self.l2_reg_w0 == 0. or self.l2_reg_w == 0.:
            logger.debug('[warn] reg_w0 and/or reg_w are fallen in 0.0')

        dot_v = np.dot(x_vec.T, self.V).reshape((self.k,))  # (k, )
        dot_prev_v = np.dot(x_vec.T, self.prev_V).reshape((self.k,))  # (k, )
        s_duplicated = np.dot((x_vec.T ** 2), self.V * self.prev_V).reshape((self.k,))  # (k, )
        self.l2_rev_V = np.maximum(np.zeros(self.k), self.l2_reg_V + coeff * (dot_v * dot_prev_v - s_duplicated))

    def update_params(self, x, value):
        x_vec = np.array([x]).T  # p x 1
        interaction = np.sum(np.dot(self.V.T, x_vec) ** 2 - np.dot(self.V.T ** 2, x_vec ** 2)) / 2.
        pred = self.w0 + np.inner(self.w, x) + interaction

        # compute current error
        err = value - pred

        self.update_reg(x, err)

        # update w0 with keeping the previous value
        self.prev_w0 = self.w0
        self.w0 = self.w0 + 2. * self.learn_rate * (err * 1. - self.l2_reg_w0 * self.w0)

        # keep the previous w for auto-parameter optimization
        self.prev_w = np.empty_like(self.w)
        self.prev_w[:] = self.w

        # keep the previous V
        self.prev_V = np.empty_like(self.V)
        self.prev_V[:] = self.V

        # update w and V
        prod = np.dot(np.array([x]), self.prev_V)  # (1, p) and (p, k) => (1, k)
        for pi in range(self.p):
            if x[pi] == 0.:
                continue

            self.w[pi] = self.prev_w[pi] + 2. * self.learn_rate * (err * x[pi] - self.l2_reg_w * self.prev_w[pi])

            g = err * x[pi] * (prod - x[pi] * self.prev_V[pi])
            self.V[pi] = self.prev_V[pi] + 2. * self.learn_rate * (g - self.l2_reg_V * self.prev_V[pi])



import numpy as np


class Base:

    def __init__(self, index, feature=np.array([0.])):
        self.index = index
        self.feature = feature

    def encode(self, dim=None,
               index=True, feature=True,
               vertical=False):

        if not dim:
            dim = self.index + 1

        x = np.array([])

        if index:
            x = np.concatenate((x, self.index_one_hot(dim)))

        if feature:
            x = np.concatenate((x, self.feature))

        return x if not vertical else np.array([x]).T

    def index_one_hot(self, dim):
        if self.index >= dim:
            raise ValueError('number of dimensions must be greater than index: %d' % self.index)

        x = np.zeros(dim)
        x[self.index] = 1.
        return x


class User(Base):
    pass


class Item(Base):
    pass


class Event:

    def __init__(self, user, item, value, context=np.array([0.])):
        self.user = user
        self.item = item
        self.value = value
        self.context = context

    def encode(self, n_user=None, n_item=None,
               index=True, feature=True, context=True,
               vertical=False):

        x = self.user.encode(dim=n_user, index=index,
                             feature=feature, vertical=False)

        if context:
            x = np.concatenate((x, self.context))

        iv = self.item.encode(dim=n_item, index=index,
                              feature=feature, vertical=False)
        x = np.concatenate((x, iv))

        return x if not vertical else np.array([x]).T



from flurs.base import BaseModel

from flurs.utils.projection import Raw
from flurs.utils.projection import RandomProjection
from flurs.utils.projection import RandomMaclaurinProjection
from flurs.utils.projection import TensorSketchProjection

import numpy as np
import numpy.linalg as ln
import scipy.sparse as sp
from sklearn import preprocessing


class OnlineSketch(BaseModel):

    """Matrix-sketching-based online item recommender

    T. Kitazawa.
    "Sketching Dynamic User-Item Interactions for Online Item Recommendation"
    In Proceedings of CHIIR 2017, March 2017.

    """

    def __init__(self, p, k=40, ell=-1, r=-1, proj='Raw'):
        # number of dimensions of input vectors
        self.p = p

        # dimension of projected vectors
        # for `Raw` (i.e. w/o projection), k must equat to p
        self.k = self.p if proj == 'Raw' else k

        # if there is no preference for ell,
        # this will be sqrt(k) similarly to what the original streaming anomaly detection paper did
        self.ell = int(np.sqrt(self.k)) if ell < 1 else ell

        # number of tracked orthogonal bases
        # * upper bound of r is ell (r <= ell) because U_r is obtained from SVD(B) (or SVD(E)), and
        #   B and E always have ell columns
        self.r = int(np.ceil(self.ell / 2)) if r < 1 else np.min(r, self.ell)

        # initialize projection instance which is specified by `proj` argument
        if proj == 'Raw':
            self.proj = Raw(self.k, self.p)
        elif proj == 'RandomProjection':
            self.proj = RandomProjection(self.k, self.p)
        elif proj == 'RandomMaclaurinProjection':
            self.proj = RandomMaclaurinProjection(self.k, self.p)
        elif proj == 'TensorSketchProjection':
            self.proj = TensorSketchProjection(self.k, self.p)

        self.init_params()

    def init_params(self):
        self.i_mat = sp.csr_matrix([])

    def update_params(self, y):
        y = self.proj.reduce(np.array([y]).T)
        y = np.ravel(preprocessing.normalize(y, norm='l2', axis=0))

        if not hasattr(self, 'B'):
            self.B = np.zeros((self.k, self.ell))

        # combine current sketched matrix with input at time t
        zero_cols = np.nonzero([np.isclose(s_col, 0.0) for s_col in np.sum(self.B, axis=0)])[0]
        j = zero_cols[0] if zero_cols.size != 0 else self.ell - 1  # left-most all-zero column in B
        self.B[:, j] = y

        U, s, V = ln.svd(self.B, full_matrices=False)

        # update the tracked orthonormal bases
        self.U_r = U[:, :self.r]

        # update ell orthogonal bases
        U_ell = U[:, :self.ell]
        s_ell = s[:self.ell]

        # shrink step in the Frequent Directions algorithm
        # (shrink singular values based on the squared smallest singular value)
        delta = s_ell[-1] ** 2
        s_ell = np.sqrt(s_ell ** 2 - delta)

        self.B = np.dot(U_ell, np.diag(s_ell))


class OnlineRandomSketch(OnlineSketch):

    """Inspired by: Streaming Anomaly Detection using Randomized Matrix Sketching
    [WIP] many matrix multiplications are computational heavy
    """

    def update_params(self, y):
        y = self.proj.reduce(np.array([y]).T)
        y = np.ravel(preprocessing.normalize(y, norm='l2', axis=0))

        if not hasattr(self, 'E'):
            self.E = np.zeros((self.k, self.ell))

        # combine current sketched matrix with input at time t
        zero_cols = np.nonzero([np.isclose(s_col, 0.0) for s_col in np.sum(self.E, axis=0)])[0]
        j = zero_cols[0] if zero_cols.size != 0 else self.ell - 1  # left-most all-zero column in B
        self.E[:, j] = y

        Gaussian = np.random.normal(0., 0.1, (self.k, 100 * self.ell))
        MM = np.dot(self.E, self.E.T)
        Q, R = ln.qr(np.dot(MM, Gaussian))

        # eig() returns eigen values/vectors with unsorted order
        s, A = ln.eig(np.dot(np.dot(Q.T, MM), Q))
        order = np.argsort(s)[::-1]
        s = s[order]
        A = A[:, order]

        U = np.dot(Q, A)

        # update the tracked orthonormal bases
        self.U_r = U[:, :self.r]

        # update ell orthogonal bases
        U_ell = U[:, :self.ell]
        s_ell = s[:self.ell]

        # shrink step in the Frequent Directions algorithm
        delta = s_ell[-1]
        s_ell = np.sqrt(s_ell - delta)

        self.E = np.dot(U_ell, np.diag(s_ell))


class OnlineSparseSketch(OnlineSketch):

    """Inspired by: Efficient Frequent Directions Algorithm for Sparse Matrices
    """

    def update_params(self, y):
        y = self.proj.reduce(np.array([y]).T)
        y = preprocessing.normalize(y, norm='l2', axis=0)  # (k, 1)

        if not hasattr(self, 'B'):
            self.p_failure = 0.1
            self.B = np.zeros((self.k, self.ell))
            self.A = np.array([])

        U, s, V = ln.svd(self.B, full_matrices=False)

        # update the tracked orthonormal bases
        self.U_r = U[:, :self.r]

        if self.A.size == 0:
            self.A = np.empty_like(y)
            self.A[:] = y[:]
        else:
            self.A = np.concatenate((self.A, y), axis=1)

        if np.count_nonzero(self.A) >= (self.ell * self.k) or self.A.shape[1] == self.k:
            B = self.__boosted_sparse_shrink(self.A, self.ell, self.p_failure)
            self.B = self.__dense_shrink(np.concatenate((self.B, B), axis=1), self.ell)
            self.A = np.array([])

    def __simultaneous_iteration(self, A, k, eps):
        n, d = A.shape

        q = int(np.log((n / eps) / eps))
        G = np.random.normal(0., 1., (d, k))

        # Gram-Schmidt
        Y = np.dot(np.dot(ln.matrix_power(np.dot(A, A.T), q), A), G)  # (n, k)
        Q, R = ln.qr(Y, mode='complete')

        return Q

    def __sparse_shrink(self, A, ell):
        m, d = A.shape

        Z = self.__simultaneous_iteration(A, ell, 1 / 4)
        P = sp.csr_matrix(np.dot(Z.T, A))

        # 1 <= ell < min(P.shape)
        H_ell, s_ell, V_ell = sp.linalg.svds(P, ell)

        # shrink step in the Frequent Directions algorithm
        # (shrink singular values based on the squared smallest singular value)
        lambda_ell = s_ell[-1] ** 2
        s_ell = np.sqrt(s_ell ** 2 - lambda_ell)

        return np.dot(H_ell, np.diag(s_ell))

    def __boosted_sparse_shrink(self, A, ell, p_failure):
        al = (6 / 41) * ell
        sq_norm_A = ln.norm(A, ord='fro') ** 2
        AA = np.dot(A, A.T)

        while True:
            B = self.__sparse_shrink(A, ell)
            delta = (sq_norm_A - (ln.norm(B, ord='fro') ** 2)) / al

            C = (AA - np.dot(B, B.T)) / (delta / 2)
            if self.__verify_spectral(C, p_failure):
                return B

    def __dense_shrink(self, A, ell):
        H, s, V = ln.svd(A, full_matrices=False)

        H_ell = H[:, :ell]
        s_ell = s[:ell]

        # shrink step in the Frequent Directions algorithm
        # (shrink singular values based on the squared smallest singular value)
        lambda_ell = s_ell[-1] ** 2
        s_ell = np.sqrt(s_ell ** 2 - lambda_ell)

        return np.dot(H_ell, np.diag(s_ell))

    def __verify_spectral(self, C, p_failure):
        d = C.shape[0]
        c = 1  # some constant

        if not hasattr(self, 'i_verify'):
            self.i_verify = 0

        self.i_verify += 1
        p_failure_i = p_failure / (2 * (self.i_verify ** 2))

        # pick a point uniformly at ranom from the d-dimensional unit sphere
        # http://stackoverflow.com/questions/15880367/python-uniform-distribution-of-points-on-4-dimensional-sphere
        normal_deviates = np.random.normal(size=(d, 1))
        radius = np.sqrt((normal_deviates ** 2).sum(axis=0))
        x = normal_deviates / radius

        return ln.norm(np.dot(ln.matrix_power(C, int(c * np.log(d / p_failure_i))), x)) <= 1



from flurs.base import BaseModel, RecommenderMixin

import numpy as np


class Popular(BaseModel, RecommenderMixin):

    """Popularity (non-personalized) baseline
    """

    def __init__(self):
        pass

    def init_params(self):
        self.freq = np.array([])

    def update_params(self, ia):
        self.freq[ia] += 1

    def init_recommender(self):
        super().init_recommender()

    def add_user(self, user):
        super().add_user(user)

    def add_item(self, item):
        super().add_item(item)
        self.freq = np.append(self.freq, 0)

    def update(self, e, is_batch_train=False):
        self.update_params(e.item.index)

    def score(self, user, candidates):
        return self.freq[candidates]

    def recommend(self, user, candidates):
        scores = self.score(user, candidates)
        return self.scores2recos(scores, candidates, rev=True)



from setuptools import setup


setup(
    name='flurs',
    version='v0.0.1',
    description='A library for online item recommendation',
    author='Takuya Kitazawa',
    author_email='k.takuti@gmail.com',
    license='MIT',
    url='https://github.com/takuti/flurs',
    download_url='https://github.com/takuti/flurs/tarball/v0.0.1',
    packages=['flurs',
              'flurs.data',
              'flurs.recommender',
              'flurs.baseline',
              'flurs.model',
              'flurs.utils'],
    install_requires=[
        'numpy',
        'scipy',
        'scikit_learn',
        'mmh3'],
)



from abc import ABCMeta, abstractmethod

import numpy as np


class BaseModel:

    """Base class for the incremental models learning from positive-only feedback.
    """
    __metaclass__ = ABCMeta

    def __init__(self, *args):
        """Set the hyperparameters.
        """
        pass

    @abstractmethod
    def init_params(self):
        """Initialize model parameters.
        """
        pass

    @abstractmethod
    def update_params(self, *args):
        """Update model parameters.
        """
        pass


class RecommenderMixin:

    """Mixin injected into a model to make it a recommender.
    """
    __metaclass__ = ABCMeta

    def init_recommender(self, *args):
        """Initialize a recommender by resetting stored users and items.
        """
        # number of observed users
        self.n_user = 0

        # store user data
        self.users = {}

        # number of observed items
        self.n_item = 0

        # store item data
        self.items = {}

    def is_new_user(self, u):
        """Check if user is new.

        Args:
            u (int): User index.

        Returns:
            boolean: Whether the user is new.

        """
        return u not in self.users

    @abstractmethod
    def add_user(self, user):
        """For new users, append their information into the dictionaries.

        Args:
            user (User): User.

        """
        self.users[user.index] = {'known_items': set()}
        self.n_user += 1

    def is_new_item(self, i):
        """Check if item is new.

        Args:
            i (int): Item index.

        Returns:
            boolean: Whether the item is new.

        """
        return i not in self.items

    @abstractmethod
    def add_item(self, item):
        """For new items, append their information into the dictionaries.

        Args:
            item (Item): Item.

        """
        self.items[item.index] = {}
        self.n_item += 1

    @abstractmethod
    def update(self, e, is_batch_train):
        """Update model parameters based on d, a sample represented as a dictionary.

        Args:
            e (Event): Observed event.

        """
        pass

    @abstractmethod
    def score(self, user, candidates):
        """Compute scores for the pairs of given user and item candidates.

        Args:
            user (User): Target user.
            candidates (numpy array; (# candidates, )): Target item' indices.

        Returns:
            numpy float array; (# candidates, ): Predicted values for the given user-candidates pairs.

        """
        return

    @abstractmethod
    def recommend(self, user, candidates):
        """Recommend items for a user represented as a dictionary d.

        First, scores are computed.
        Next, `self.__scores2recos()` is called to convert the scores into a recommendation list.

        Args:
            user (User): Target user.
            candidates (numpy array; (# target items, )): Target items' indices. Only these items are considered as the recommendation candidates.

        Returns:
            (numpy array, numpy array) : (Sorted list of items, Sorted scores).

        """
        return

    def scores2recos(self, scores, candidates, rev=False):
        """Get recommendation list for a user u_index based on scores.

        Args:
            scores (numpy array; (n_target_items,)):
                Scores for the target items. Smaller score indicates a promising item.
            candidates (numpy array; (# target items, )): Target items' indices. Only these items are considered as the recommendation candidates.
            rev (bool): If true, return items in an descending order. A ascending order (i.e., smaller scores are more promising) is default.

        Returns:
            (numpy array, numpy array) : (Sorted list of items, Sorted scores).

        """
        sorted_indices = np.argsort(scores)

        if rev:
            sorted_indices = sorted_indices[::-1]

        return candidates[sorted_indices], scores[sorted_indices]


class FeatureRecommenderMixin(RecommenderMixin):

    """Mixin injected into a model to make it a feature-based recommender.
    """
    __metaclass__ = ABCMeta

    @abstractmethod
    def score(self, user, candidates, context):
        """Compute scores for the pairs of given user and item candidates.

        Args:
            user (User): Target user.
            candidates (numpy array; (# candidates, )): Target item' indices.
            context (numpy 1d array): Feature vector representing contextual information.

        Returns:
            numpy float array; (# candidates, ): Predicted values for the given user-candidates pairs.

        """
        return

    @abstractmethod
    def recommend(self, user, candidates, context):
        """Recommend items for a user represented as a dictionary d.

        First, scores are computed.
        Next, `self.__scores2recos()` is called to convert the scores into a recommendation list.

        Args:
            user (User): Target user.
            candidates (numpy array; (# target items, )): Target items' indices. Only these items are considered as the recommendation candidates.
            context (numpy 1d array): Feature vector representing contextual information.

        Returns:
            (numpy array, numpy array) : (Sorted list of items, Sorted scores).

        """
        return



from flurs.base import BaseModel

import numpy as np


class UserKNN(BaseModel):

    """Incremental User-based Collaborative Filtering

    M. Pepagelis et al.
    "Incremental Collaborative Filtering for Highly-Scalable Recommendation Algorithms"
    In Foundations of Intelligent Systems, pages 553-561. Springer Berlin Heidelberg, 2005.

    """

    def __init__(self, k=5):
        # number of nearest neighbors
        self.k = k

        self.init_params()

    def init_params(self):
        # user-item matrix
        self.R = np.array([])

        # user-user similarity matrix
        self.S = np.array([])

        # user-user similarity: S = B / (sqrt(C) * sqrt(D))
        self.B = np.array([])
        self.C = np.array([])
        self.D = np.array([])

    def update_params(self, ua, ia, value):
        prev_r = self.R[ua, ia]
        is_new_submit = (prev_r == 0)
        self.R[ua, ia] = value

        prev_mean = self.users[ua]['mean']

        if is_new_submit:
            self.users[ua]['count'] += 1
            self.users[ua]['mean'] = self.R[ua, ia] / self.users[ua]['count'] + (
                self.users[ua]['count'] - 1) / self.users[ua]['count'] * prev_mean
        else:
            self.users[ua]['mean'] = (self.R[ua, ia] - prev_r) / (self.users[ua]['count'] - 1) + prev_mean

        d = self.users[ua]['mean'] - prev_mean

        for uy in range(self.n_user):
            # skip myself
            if uy == ua:
                continue

            e = f = g = 0.

            had_uy_rated_ia = (self.R[uy, ia] != 0)
            if had_uy_rated_ia:
                ua_normalized = self.R[ua, ia] - self.users[ua]['mean']
                uy_normalized = self.R[uy, ia] - self.users[uy]['mean']

                if is_new_submit:
                    e = ua_normalized * uy_normalized
                    f = ua_normalized ** 2
                    g = uy_normalized ** 2
                else:
                    e = (self.R[ua, ia] - prev_r) * uy_normalized
                    f = (self.R[ua, ia] - prev_r) ** 2 + 2 * (self.R[ua, ia] - prev_r) * ua_normalized
                    g = 0.

            for ih in range(self.n_item):
                # only for co-rated items
                if self.R[ua, ih] != 0 and self.R[uy, ih] != 0:
                    e = e - d * (self.R[uy, ih] - self.users[uy]['mean'])
                    f = f + d ** 2 - 2 * d * (self.R[ua, ih] - prev_mean)

            self.B[ua, uy] += e
            self.C[ua, uy] += f
            self.D[ua, uy] += g

        # avoid zero division
        idx = (self.C[ua, :] != 0) & (self.C[ua, :] != 0)
        self.S[ua, idx] = self.B[ua, idx] / (np.sqrt(self.C[ua, idx]) * np.sqrt(self.D[ua, idx]))
